@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2018 Artifex Software, Inc.
+/* Copyright (C) 2001-2021 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -31,6 +31,7 @@
 #include "gsstype.h"
 #include "gdevprn.h"
 #include "gdevp14.h"        /* Needed to patch up the procs after compositor creation */
+#include "gximage.h"        /* For gx_image_enum */
 #include "gdevsclass.h"
 #include "gdevoflt.h"
 
@@ -66,6 +67,7 @@ static dev_proc_fill_linear_color_triangle(obj_filter_fill_linear_color_triangle
 static dev_proc_put_image(obj_filter_put_image);
 static dev_proc_strip_copy_rop2(obj_filter_strip_copy_rop2);
 static dev_proc_strip_tile_rect_devn(obj_filter_strip_tile_rect_devn);
+static dev_proc_fill_stroke_path(obj_filter_fill_stroke_path);
 
 /* The device prototype */
 #define MAX_COORD (max_int_in_fixed - 1000)
@@ -73,7 +75,7 @@ static dev_proc_strip_tile_rect_devn(obj_filter_strip_tile_rect_devn);
 
 #define public_st_obj_filter_device()	/* in gsdevice.c */\
   gs_public_st_complex_only(st_obj_filter_device, gx_device, "object filter",\
-    0, obj_filter_enum_ptrs, obj_filter_reloc_ptrs, gx_device_finalize)
+    0, obj_filter_enum_ptrs, obj_filter_reloc_ptrs, default_subclass_finalize)
 
 static
 ENUM_PTRS_WITH(obj_filter_enum_ptrs, gx_device *dev);
@@ -172,7 +174,9 @@ gx_device_obj_filter gs_obj_filter_device =
      obj_filter_strip_copy_rop2,
      obj_filter_strip_tile_rect_devn,
      default_subclass_copy_alpha_hl_color,
-     default_subclass_process_page
+     default_subclass_process_page,
+     default_subclass_transform_pixel_region,
+     obj_filter_fill_stroke_path,
     }
 };
 
@@ -220,6 +224,16 @@ int obj_filter_stroke_path(gx_device *dev, const gs_gstate *pgs, gx_path *ppath,
     return 0;
 }
 
+int obj_filter_fill_stroke_path(gx_device *dev, const gs_gstate *pgs, gx_path *ppath,
+        const gx_fill_params *fill_params, const gx_drawing_color *pdcolor_fill,
+        const gx_stroke_params *stroke_params, const gx_drawing_color *pdcolor_stroke,
+        const gx_clip_path *pcpath)
+{
+    if ((dev->ObjectFilter & FILTERVECTOR) == 0)
+        return default_subclass_fill_stroke_path(dev, pgs, ppath, fill_params, pdcolor_fill, stroke_params, pdcolor_stroke, pcpath);
+    return 0;
+}
+
 int obj_filter_fill_mask(gx_device *dev, const byte *data, int data_x, int raster, gx_bitmap_id id,
     int x, int y, int width, int height,
     const gx_drawing_color *pdcolor, int depth,
@@ -311,6 +325,8 @@ int obj_filter_strip_copy_rop(gx_device *dev, const byte *sdata, int sourcex, ui
 
 typedef struct obj_filter_image_enum_s {
     gx_image_enum_common;
+    int y;
+    int height;
 } obj_filter_image_enum;
 gs_private_st_composite(st_obj_filter_image_enum, obj_filter_image_enum, "obj_filter_image_enum",
   obj_filter_image_enum_enum_ptrs, obj_filter_image_enum_reloc_ptrs);
@@ -331,7 +347,17 @@ obj_filter_image_plane_data(gx_image_enum_common_t * info,
                      const gx_image_plane_t * planes, int height,
                      int *rows_used)
 {
-    return 0;
+    obj_filter_image_enum *pie = (obj_filter_image_enum *)info;
+
+    if (height > pie->height - pie->y)
+        height = pie->height - pie->y;
+
+    pie->y += height;
+    *rows_used = height;
+
+    if (pie->y < pie->height)
+        return 0;
+    return 1;
 }
 
 static int
@@ -378,6 +404,8 @@ int obj_filter_begin_typed_image(gx_device *dev, const gs_gstate *pgs, const gs_
                         (gx_device *)dev, num_components, pim->format);
     pie->memory = memory;
     pie->skipping = true;
+    pie->height = pim->Height;
+    pie->y = 0;
 
     return 0;
 }
@@ -518,12 +546,12 @@ int obj_filter_fill_linear_color_triangle(gx_device *dev, const gs_fill_attribut
     return 0;
 }
 
-int obj_filter_put_image(gx_device *dev, const byte **buffers, int num_chan, int x, int y,
+int obj_filter_put_image(gx_device *dev, gx_device *mdev, const byte **buffers, int num_chan, int x, int y,
             int width, int height, int row_stride,
             int alpha_plane_index, int tag_plane_index)
 {
     if ((dev->ObjectFilter & FILTERIMAGE) == 0)
-        return default_subclass_put_image(dev, buffers, num_chan, x, y, width, height, row_stride, alpha_plane_index, tag_plane_index);
+        return default_subclass_put_image(dev, mdev, buffers, num_chan, x, y, width, height, row_stride, alpha_plane_index, tag_plane_index);
     return 0;
 }
 
